from django.utils.encoding import force_str

from mayan.apps.credentials.tests.mixins import StoredCredentialPasswordUsernameTestMixin
from mayan.apps.source_compressed.source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_NEVER
from mayan.apps.source_periodic.source_backends.literals import DEFAULT_PERIOD_INTERVAL
from mayan.apps.sources.tests.literals import (
    TEST_CASE_ACTION_NAME_SOURCE_CREATE, TEST_CASE_INTERFACE_NAME_VIEW
)
from mayan.apps.sources.tests.mixins.base_mixins import SourceTestMixin

from ..source_backends.literals import (
    DEFAULT_EMAIL_IMAP_MAILBOX, DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
    DEFAULT_EMAIL_IMAP_STORE_COMMANDS, DEFAULT_EMAIL_POP3_TIMEOUT
)

from .literals import (
    TEST_EMAIL_ATTACHMENT_AND_INLINE, TEST_EMAIL_SOURCE_PASSWORD,
    TEST_EMAIL_SOURCE_USERNAME, TEST_SOURCE_BACKEND_PATH_TEST_EMAIL,
    TEST_SOURCE_BACKEND_PATH_TEST_EMAIL_IMAP,
    TEST_SOURCE_BACKEND_PATH_TEST_EMAIL_POP3
)


class CredentialSourceTestMixin(StoredCredentialPasswordUsernameTestMixin):
    _test_stored_credential_backend_data = {
        'password': TEST_EMAIL_SOURCE_PASSWORD,
        'username': TEST_EMAIL_SOURCE_USERNAME
    }


class EmailSourceTestMixin(SourceTestMixin, CredentialSourceTestMixin):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_TEST_EMAIL
    _test_source_content = TEST_EMAIL_ATTACHMENT_AND_INLINE

    def _test_source_create(self, *args, **kwargs):
        super()._test_source_create(*args, **kwargs)
        backend_instance = self._test_source.get_backend_instance()
        backend_instance.reset_test_mock_server(
            test_content=self._test_source_content
        )

    def _get_test_source_backend_data(self, action_name, interface_name):
        result = super()._get_test_source_backend_data(
            action_name=action_name, interface_name=interface_name
        )

        result['_test_content'] = force_str(s=self._test_source_content)

        if action_name == TEST_CASE_ACTION_NAME_SOURCE_CREATE:
            result.update(
                {
                    'document_type_id': self._test_document_type.pk,
                    'host': '',
                    'interval': DEFAULT_PERIOD_INTERVAL,
                    'port': '',
                    'ssl': False,
                    'store_body': False,
                    'stored_credential_id': self._test_stored_credential.pk,
                    'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER
                }
            )

            if interface_name == TEST_CASE_INTERFACE_NAME_VIEW:
                result.update(
                    {
                        'enabled': True,
                        'host': 'fake-host',
                        'port': 0,
                        'stored_credential_id': self._test_stored_credential.pk
                    }
                )

        return result

    def get_test_source_stored_file_list(self):
        backend_instance = self._test_source.get_backend_instance()

        return list(
            backend_instance.get_stored_file_list()
        )


class IMAPEmailSourceTestMixin(EmailSourceTestMixin):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_TEST_EMAIL_IMAP

    def _get_test_source_backend_data(self, action_name, interface_name):
        result = super()._get_test_source_backend_data(
            action_name=action_name, interface_name=interface_name
        )

        if action_name == TEST_CASE_ACTION_NAME_SOURCE_CREATE:
            result.update(
                {
                    'execute_expunge': True,
                    'mailbox': DEFAULT_EMAIL_IMAP_MAILBOX,
                    'mailbox_destination': '',
                    'port': 143,
                    'search_criteria': DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
                    'store_commands': DEFAULT_EMAIL_IMAP_STORE_COMMANDS,
                }
            )

        return result


class POP3EmailSourceTestMixin(EmailSourceTestMixin):
    _test_source_backend_path = TEST_SOURCE_BACKEND_PATH_TEST_EMAIL_POP3

    def _get_test_source_backend_data(self, action_name, interface_name):
        result = super()._get_test_source_backend_data(
            action_name=action_name, interface_name=interface_name
        )

        if action_name == TEST_CASE_ACTION_NAME_SOURCE_CREATE:
            result.update(
                {
                    'port': 110,
                    'timeout': DEFAULT_EMAIL_POP3_TIMEOUT
                }
            )

        return result



from django.apps import apps
from django.db.models.signals import post_migrate
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_list_facet, menu_object, menu_secondary
)
from mayan.apps.databases.classes import ModelFieldRelated
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.rest_api.fields import DynamicSerializerField

from .events import (
    event_signature_capture_created, event_signature_capture_deleted,
    event_signature_capture_edited
)
from .handlers import handler_signature_capture_cache_create
from .links import (
    link_signature_capture_create, link_signature_capture_delete,
    link_signature_capture_edit, link_signature_capture_list
)
from .permissions import (
    permission_signature_capture_create, permission_signature_capture_delete,
    permission_signature_capture_edit, permission_signature_capture_view
)
from .transformations import *  # NOQA


class SignatureCapturesApp(MayanAppConfig):
    app_namespace = 'signature_captures'
    app_url = 'signature_captures'
    has_rest_api = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.signature_captures'
    verbose_name = _(message='Signature captures')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        SignatureCapture = self.get_model(
            model_name='SignatureCapture'
        )

        DynamicSerializerField.add_serializer(
            klass=SignatureCapture,
            serializer_class='mayan.apps.signature_captures.serializers.SignatureCapture'
        )

        EventModelRegistry.register(model=SignatureCapture)

        ModelEventType.register(
            model=Document, event_types=(
                event_signature_capture_created,
                event_signature_capture_deleted,
                event_signature_capture_edited
            )
        )
        ModelEventType.register(
            model=SignatureCapture, event_types=(
                event_signature_capture_edited,
            )
        )

        ModelFieldRelated(
            model=Document, name='signature_captures__text'
        )
        ModelFieldRelated(
            model=Document, name='signature_captures__internal_name'
        )
        ModelFieldRelated(
            model=Document, name='signature_captures__user__first_name'
        )
        ModelFieldRelated(
            model=Document, name='signature_captures__user__last_name'
        )
        ModelFieldRelated(
            model=Document, name='signature_captures__user__username'
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_signature_capture_create,
                permission_signature_capture_delete,
                permission_signature_capture_view
            )
        )

        ModelPermission.register(
            model=SignatureCapture, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_signature_capture_delete,
                permission_signature_capture_edit,
                permission_signature_capture_view
            )
        )

        ModelPermission.register_inheritance(
            model=SignatureCapture, related='document'
        )

        # SignatureCapture

        SourceColumn(
            attribute='date_time_created', is_identifier=True,
            is_sortable=True, source=SignatureCapture
        )
        SourceColumn(
            attribute='date_time_edited', is_sortable=True,
            source=SignatureCapture
        )
        SourceColumn(
            attribute='text', is_sortable=True,
            source=SignatureCapture
        )
        SourceColumn(
            attribute='internal_name', is_sortable=True,
            source=SignatureCapture
        )
        SourceColumn(
            attribute='user', is_sortable=True,
            source=SignatureCapture
        )

        # Document

        menu_list_facet.bind_links(
            links=(link_signature_capture_list,), sources=(Document,)
        )

        menu_object.bind_links(
            links=(
                link_signature_capture_edit,
                link_signature_capture_delete
            ),
            sources=(SignatureCapture,)
        )

        menu_secondary.bind_links(
            links=(
                link_signature_capture_create, link_signature_capture_list
            ),
            sources=(
                SignatureCapture,
                'signature_captures:signature_capture_create',
                'signature_captures:signature_capture_list'
            )
        )

        post_migrate.connect(
            dispatch_uid='signature_captures_handler_signature_capture_cache_create',
            receiver=handler_signature_capture_cache_create
        )



import logging

from django.apps import apps

from mayan.celery import app

logger = logging.getLogger(name=__name__)


@app.task(ignore_result=True)
def task_remove_metadata_type(document_type_id, metadata_type_id):
    DocumentMetadata = apps.get_model(
        app_label='metadata', model_name='DocumentMetadata'
    )

    DocumentMetadata.objects.filter(
        document__document_type__id=document_type_id,
        metadata_type__id=metadata_type_id
    ).delete()


@app.task(ignore_result=True)
def task_add_required_metadata_type(document_type_id, metadata_type_id):
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )

    MetadataType = apps.get_model(
        app_label='metadata', model_name='MetadataType'
    )

    metadata_type = MetadataType.objects.get(pk=metadata_type_id)

    for document in DocumentType.objects.get(pk=document_type_id).documents.all():
        document.metadata.create(metadata_type=metadata_type)



from django.utils.translation import gettext_lazy as _

from mayan.apps.sources.source_backend_actions.interface_arguments import SourceBackendActionInterfaceArgument

argument_file_object = SourceBackendActionInterfaceArgument(
    help_text=_(message='Python file like object.')
)
argument_shared_uploaded_file_id = SourceBackendActionInterfaceArgument(
    help_text=_(
        'ID of the shared uploaded file to be processed.'
    )
)



from django.shortcuts import get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.rest_api.api_view_mixins import AsymmetricSerializerAPIViewMixin

from ..models.index_template_models import IndexTemplate
from ..permissions import (
    permission_index_template_edit, permission_index_template_view
)
from ..serializers import (
    IndexTemplateNodeSerializer, IndexTemplateNodeWriteSerializer
)


class APIIndexTemplateNodeViewMixin(AsymmetricSerializerAPIViewMixin):
    object_permission_map = {
        'DELETE': permission_index_template_edit,
        'GET': permission_index_template_view,
        'PATCH': permission_index_template_edit,
        'PUT': permission_index_template_edit,
        'POST': permission_index_template_edit
    }
    read_serializer_class = IndexTemplateNodeSerializer
    write_serializer_class = IndexTemplateNodeWriteSerializer

    def get_serializer(self, *args, **kwargs):
        if not self.request:
            return None

        return super().get_serializer(*args, **kwargs)

    def get_index_template(self):
        if 'index_template_id' in self.kwargs:
            permission = self.object_permission_map[self.request.method]

            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=IndexTemplate.objects.all(),
                user=self.request.user
            )

            return get_object_or_404(
                klass=queryset, pk=self.kwargs['index_template_id']
            )

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['index_template'] = self.get_index_template()
        return context



from mayan.apps.source_web_forms.tests.mixins import WebFormSourceTestMixin
from mayan.apps.sources.tests.mixins.source_view_mixins import SourceActionViewTestMixin
from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..models import Cabinet

from .literals import (
    TEST_CABINET_CHILD_LABEL, TEST_CABINET_LABEL, TEST_CABINET_LABEL_EDITED
)


class CabinetTestMixin(TestMixinObjectCreationTrack):
    _test_cabinet_add_test_document = False
    _test_object_model = Cabinet
    _test_object_name = '_test_cabinet'
    auto_create_test_cabinet = False

    def setUp(self):
        super().setUp()

        self._test_cabinet_list = []

        if self.auto_create_test_cabinet:
            self._create_test_cabinet(
                add_test_document=self._test_cabinet_add_test_document
            )

    def _create_test_cabinet(self, label=None, add_test_document=False):
        label = self._get_test_cabinet_label(label=label)

        self._test_cabinet = Cabinet.objects.create(label=label)

        if add_test_document:
            self._test_cabinet._document_add(document=self._test_document)

        self._test_cabinet_list.append(self._test_cabinet)

    def _create_test_cabinet_child(self, label=None):
        self._test_cabinet_child = Cabinet.objects.create(
            label=label or TEST_CABINET_CHILD_LABEL,
            parent=self._test_cabinet
        )

    def _get_test_cabinet_label(self, label=None):
        if label is None:
            test_cabinet_count = len(self._test_cabinet_list)
            label = '{}_{}'.format(TEST_CABINET_LABEL, test_cabinet_count)

        return label


class CabinetAPIViewTestMixin(CabinetTestMixin):
    def _request_test_cabinet_child_create_api_view(self):
        data = {
            'label': TEST_CABINET_CHILD_LABEL,
            'parent': self._test_cabinet.pk
        }

        self._test_object_track()

        response = self.post(viewname='rest_api:cabinet-list', data=data)

        self._test_object_set(test_object_name='_test_cabinet_child')

        return response

    def _request_test_cabinet_child_delete_api_view(self):
        return self.delete(
            viewname='rest_api:cabinet-detail', kwargs={
                'cabinet_id': self._test_cabinet_child.pk
            }
        )

    def _request_test_cabinet_create_api_view(
        self, extra_data=None, label=None
    ):
        label = self._get_test_cabinet_label(label=label)

        data = {'label': label, 'parent': ''}

        if extra_data:
            data.update(extra_data)

        self._test_object_track()

        response = self.post(viewname='rest_api:cabinet-list', data=data)

        self._test_object_set()

        return response

    def _request_test_cabinet_delete_api_view(self):
        return self.delete(
            viewname='rest_api:cabinet-detail', kwargs={
                'cabinet_id': self._test_cabinet.pk
            }
        )

    def _request_test_cabinet_edit_api_patch_view(self):
        return self.patch(
            data={'label': TEST_CABINET_LABEL_EDITED}, kwargs={
                'cabinet_id': self._test_cabinet.pk
            }, viewname='rest_api:cabinet-detail'
        )

    def _request_test_cabinet_edit_api_put_view(self):
        return self.put(
            data={'label': TEST_CABINET_LABEL_EDITED}, kwargs={
                'cabinet_id': self._test_cabinet.pk
            }, viewname='rest_api:cabinet-detail'
        )

    def _request_test_cabinet_list_api_view(self):
        return self.get(viewname='rest_api:cabinet-list')


class CabinetDocumentAPIViewTestMixin(CabinetTestMixin):
    def _request_test_cabinet_document_add_api_view(self):
        return self.post(
            viewname='rest_api:cabinet-document-add', kwargs={
                'cabinet_id': self._test_cabinet.pk
            }, data={
                'document': self._test_document.pk
            }
        )

    def _request_test_cabinet_document_list_api_view(self):
        return self.get(
            viewname='rest_api:cabinet-document-list', kwargs={
                'cabinet_id': self._test_cabinet.pk
            }
        )

    def _request_test_cabinet_document_remove_api_view(self):
        return self.post(
            viewname='rest_api:cabinet-document-remove', kwargs={
                'cabinet_id': self._test_cabinet.pk
            }, data={
                'document': self._test_document.pk
            }
        )


class CabinetDocumentUploadWizardStepTestMixin(
    CabinetTestMixin, WebFormSourceTestMixin, SourceActionViewTestMixin
):
    def _request_test_source_document_upload_post_view_with_cabinets(self):
        cabinet_id_list = list(
            Cabinet.objects.values_list('pk', flat=True)
        )

        return self._request_test_source_document_upload_post_view(
            extra_data={'cabinets': cabinet_id_list}
        )


class CabinetViewTestMixin(CabinetTestMixin):
    def _request_test_cabinet_create_view(self, label=None):
        self._test_object_track()

        label = self._get_test_cabinet_label(label=label)

        response = self.post(
            'cabinets:cabinet_create', data={'label': label}
        )

        self._test_object_set()

        self._test_cabinet_list.append(self._test_cabinet)

        return response

    def _request_test_cabinet_delete_view(self):
        return self.post(
            viewname='cabinets:cabinet_delete', kwargs={
                'cabinet_id': self._test_cabinet.pk
            }
        )

    def _request_test_cabinet_edit_view(self):
        return self.post(
            viewname='cabinets:cabinet_edit', kwargs={
                'cabinet_id': self._test_cabinet.pk
            }, data={
                'label': TEST_CABINET_LABEL_EDITED
            }
        )

    def _request_test_cabinet_child_create_view(self):
        self._test_object_track()

        response = self.post(
            viewname='cabinets:cabinet_child_add', kwargs={
                'cabinet_id': self._test_cabinet.pk
            }, data={'label': TEST_CABINET_CHILD_LABEL}
        )

        self._test_object_set(test_object_name='_test_cabinet_child')

        return response

    def _request_test_cabinet_child_delete_view(self):
        return self.post(
            viewname='cabinets:cabinet_delete', kwargs={
                'cabinet_id': self._test_cabinet_child.pk
            }
        )

    def _request_test_cabinet_document_list_view(self):
        return self.get(
            viewname='cabinets:cabinet_view', kwargs={
                'cabinet_id': self._test_cabinet.pk
            }
        )

    def _request_test_cabinet_list_view(self):
        return self.get(viewname='cabinets:cabinet_list')

    def _request_test_document_cabinet_add_view(self):
        return self.post(
            viewname='cabinets:document_cabinet_add', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'cabinets': self._test_cabinet.pk
            }
        )

    def _request_test_document_cabinet_multiple_remove_view(self):
        return self.post(
            viewname='cabinets:document_cabinet_remove', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'cabinets': (self._test_cabinet.pk,)
            }
        )

    def _request_test_document_multiple_cabinet_multiple_add_view_cabinet(
        self
    ):
        return self.post(
            viewname='cabinets:document_multiple_cabinet_add', data={
                'id_list': (self._test_document.pk,),
                'cabinets': self._test_cabinet.pk
            }
        )


class DocumentCabinetAPIViewTestMixin(CabinetTestMixin):
    def _request_test_document_cabinet_list_api_view(self):
        return self.get(
            viewname='rest_api:document-cabinet-list', kwargs={
                'document_id': self._test_document.pk
            }
        )


class DocumentCabinetViewTestMixin(CabinetTestMixin):
    def _request_test_document_cabinet_list_view(self):
        return self.get(
            viewname='cabinets:document_cabinet_list', kwargs={
                'document_id': self._test_document.pk
            }
        )



from mayan.apps.sources.source_backend_actions.interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceTask
)

from .arguments import argument_dry_run


class SourceBackendActionMixinPeriodic:
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                dry_run = argument_dry_run

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['dry_run'] = self.context['dry_run']

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                dry_run = argument_dry_run

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['dry_run'] = self.context['dry_run']



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_source_created, event_source_edited
from ..models import Source
from ..permissions import (
    permission_sources_create, permission_sources_delete,
    permission_sources_edit, permission_sources_view
)

from .literals import (
    TEST_SOURCE_ACTION_CONFIRM_FALSE_NAME,
    TEST_SOURCE_ACTION_CONFIRM_TRUE_NAME
)
from .mixins.source_api_view_mixins import (
    SourceActionAPIViewTestMixin, SourceAPIViewTestMixin
)


class SourceActionAPIViewTestCase(
    SourceActionAPIViewTestMixin, BaseAPITestCase
):
    def test_source_action_detail_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_source_action_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_action_detail_api_view_with_access(self):
        self.grant_access(
            obj=self._test_source, permission=permission_sources_view
        )

        self._clear_events()

        response = self._request_test_source_action_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        test_source_action = list(self._test_source.actions)[0]
        self.assertEqual(
            response.data['accept_files'], test_source_action.accept_files
        )
        self.assertEqual(
            response.data['confirmation'], test_source_action.confirmation
        )
        self.assertEqual(
            response.data['name'], test_source_action.name
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_action_execute_get_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_source_action_execute_get_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_action_execute_get_api_view_with_access(self):
        action = self._test_source.get_action(
            name=TEST_SOURCE_ACTION_CONFIRM_FALSE_NAME
        )

        self.grant_access(
            obj=self._test_source, permission=action.permission
        )

        self._clear_events()

        response = self._request_test_source_action_execute_get_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_action_execute_post_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_source_action_execute_post_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_action_execute_post_api_view_with_access(self):
        action = self._test_source.get_action(
            name=TEST_SOURCE_ACTION_CONFIRM_TRUE_NAME
        )

        self.grant_access(
            obj=self._test_source, permission=action.permission
        )

        self._clear_events()

        response = self._request_test_source_action_execute_post_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_action_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_source_action_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_action_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_source, permission=permission_sources_view
        )

        self._clear_events()

        response = self._request_test_source_action_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        test_source_action_list = list(
            self._test_source.get_action_list()
        )

        self.assertEqual(
            response.data['count'], 2
        )
        self.assertEqual(
            response.data['results'][0]['name'],
            test_source_action_list[0].name
        )
        self.assertEqual(
            response.data['results'][1]['name'],
            test_source_action_list[1].name
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class SourceAPIViewTestCase(SourceAPIViewTestMixin, BaseAPITestCase):
    _test_source_create_auto = False

    def test_source_create_api_view_no_permission(self):
        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_source)
        self.assertEqual(events[0].verb, event_source_created.id)

    def test_source_delete_api_view_no_permission(self):
        self._test_source_create()

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_delete_api_view_with_access(self):
        self._test_source_create()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_delete
        )

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_source_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Source.objects.count(), source_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_detail_api_view_no_permission(self):
        self._test_source_create()

        self._clear_events()

        response = self._request_test_source_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_detail_api_view_with_access(self):
        self._test_source_create()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_view
        )

        self._clear_events()

        response = self._request_test_source_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['id'], self._test_source.pk)
        self.assertEqual(response.data['label'], self._test_source.label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_edit_api_view_via_patch_no_permission(self):
        self._test_source_create()

        source_label = self._test_source.label

        self._clear_events()

        response = self._request_test_source_edit_api_view_via_patch()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_source.refresh_from_db()
        self.assertEqual(self._test_source.label, source_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_edit_api_view_via_patch_with_access(self):
        self._test_source_create()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        source_label = self._test_source.label

        self._clear_events()

        response = self._request_test_source_edit_api_view_via_patch()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_source.refresh_from_db()
        self.assertNotEqual(self._test_source.label, source_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_source)
        self.assertEqual(events[0].verb, event_source_edited.id)

    def test_source_edit_api_view_via_put_no_permission(self):
        self._test_source_create()

        source_label = self._test_source.label

        self._clear_events()

        response = self._request_test_source_edit_api_view_via_put()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_source.refresh_from_db()
        self.assertEqual(self._test_source.label, source_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_edit_api_view_via_put_with_access(self):
        self._test_source_create()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        source_label = self._test_source.label

        self._clear_events()

        response = self._request_test_source_edit_api_view_via_put()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_source.refresh_from_db()
        self.assertNotEqual(self._test_source.label, source_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_source)
        self.assertEqual(events[0].verb, event_source_edited.id)

    def test_source_list_api_view_no_permission(self):
        self._test_source_create()

        self._clear_events()

        response = self._request_test_source_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_list_api_view_with_access(self):
        self._test_source_create()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_view
        )
        self._clear_events()

        response = self._request_test_source_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['label'],
            self._test_source.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Common'), name='common'
)

permission_object_copy = namespace.add_permission(
    label=_(message='Copy object'), name='object_copy'
)



from django.utils.module_loading import import_string


def execute_callback(callback_dict, name, **kwargs):
    callback_info = callback_dict.pop(
        name, {}
    )

    if callback_info:
        callback_obj = import_string(
            dotted_path=callback_info['dotted_path']
        )
        callback_kwargs = callback_info.get(
            'kwargs', {}
        )
        function = getattr(
            callback_obj, callback_info['function_name']
        )
        function(**callback_kwargs, **kwargs)



from django.http import HttpResponseRedirect

from mayan.apps.templating.classes import Template

from .events import event_web_link_edited, event_web_link_navigated


class ResolvedWebLinkBusinessLogicMixin:
    def get_redirect(self, document, user):
        event_web_link_navigated.commit(
            actor=user, action_object=document,
            target=self
        )
        return HttpResponseRedirect(
            redirect_to=self.get_redirect_url_for(document=document)
        )

    def get_redirect_url_for(self, document):
        return Template(template_string=self.template).render(
            context={'document': document}
        )


class WebLinkBusinessLogicMixin:
    def document_types_add(self, queryset, user=None):
        for document_type in queryset:
            self.document_types.add(document_type)
            event_web_link_edited.commit(
                action_object=document_type, actor=user, target=self
            )

    def document_types_remove(self, queryset, user=None):
        for document_type in queryset:
            self.document_types.remove(document_type)
            event_web_link_edited.commit(
                action_object=document_type, actor=user, target=self
            )



from django.urls import re_path

from .api_views import (
    APIDocumentTagAttachView, APIDocumentTagListView,
    APIDocumentTagRemoveView, APITagDetailView, APITagDocumentListView,
    APITagListView
)
from .views import (
    DocumentTagListView, TagAttachActionView, TagCreateView, TagDeleteView,
    TagDocumentListView, TagEditView, TagListView, TagRemoveActionView
)

urlpatterns_documents = [
    re_path(
        route=r'^documents/(?P<document_id>\d+)/tags/$',
        name='document_tag_list', view=DocumentTagListView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>\d+)/tags/multiple/attach/$',
        name='tag_attach', view=TagAttachActionView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>\d+)/tags/multiple/remove/$',
        name='single_document_multiple_tag_remove',
        view=TagRemoveActionView.as_view()
    ),
    re_path(
        route=r'^documents/multiple/tags/multiple/remove/$',
        name='multiple_documents_selection_tag_remove',
        view=TagRemoveActionView.as_view()
    ),
    re_path(
        route=r'^documents/multiple/tags/multiple/attach/$',
        name='multiple_documents_tag_attach',
        view=TagAttachActionView.as_view()
    )
]

urlpatterns_tags = [
    re_path(
        route=r'^tags/$', name='tag_list', view=TagListView.as_view()
    ),
    re_path(
        route=r'^tags/create/$', name='tag_create',
        view=TagCreateView.as_view()
    ),
    re_path(
        route=r'^tags/(?P<tag_id>\d+)/delete/$', name='tag_single_delete',
        view=TagDeleteView.as_view()
    ),
    re_path(
        route=r'^tags/(?P<tag_id>\d+)/edit/$', name='tag_edit',
        view=TagEditView.as_view()
    ),
    re_path(
        route=r'^tags/(?P<tag_id>\d+)/documents/$', name='tag_document_list',
        view=TagDocumentListView.as_view()
    ),
    re_path(
        route=r'^tags/multiple/delete/$', name='tag_multiple_delete',
        view=TagDeleteView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_documents)
urlpatterns.extend(urlpatterns_tags)

api_urls = [
    re_path(
        route=r'^tags/$', view=APITagListView.as_view(), name='tag-list'
    ),
    re_path(
        route=r'^tags/(?P<tag_id>[0-9]+)/$', view=APITagDetailView.as_view(),
        name='tag-detail'
    ),
    re_path(
        route=r'^tags/(?P<tag_id>[0-9]+)/documents/$',
        view=APITagDocumentListView.as_view(), name='tag-document-list'
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/tags/$',
        view=APIDocumentTagListView.as_view(), name='document-tag-list'
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/tags/attach/$',
        name='document-tag-attach', view=APIDocumentTagAttachView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/tags/remove/$',
        name='document-tag-remove', view=APIDocumentTagRemoveView.as_view()
    )
]



import logging

from celery import chord

from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import OperationalError

from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

from .classes import FileMetadataDriver
from .events import event_file_metadata_document_file_finished
from .literals import LOCK_EXPIRE

logger = logging.getLogger(name=__name__)


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_metadata_process(
    self, document_file_id, user_id=None
):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )
    try:
        document_file = DocumentFile.objects.get(pk=document_file_id)
    except OperationalError as exception:
        raise self.retry(exc=exception)
    else:
        driver_class_list = FileMetadataDriver.collection.get_driver_for_mime_type(
            mime_type=document_file.mimetype
        )

        document_file_metadata_driver_task_list = []
        for driver_class in driver_class_list:
            document_file_metadata_driver_task_list.append(
                task_metadata_driver_process.s(
                    document_file_id=document_file_id,
                    stored_driver_id=driver_class.model_instance.pk
                )
            )

        chord(document_file_metadata_driver_task_list)(
            task_document_file_metadata_finished.s(
                document_file_id=document_file_id, user_id=user_id
            )
        )


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_metadata_finished(
    self, results, document_file_id, user_id=None
):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    User = get_user_model()

    try:
        document_file = DocumentFile.objects.get(pk=document_file_id)

        if user_id:
            user = User.objects.get(pk=user_id)
        else:
            user = None
    except OperationalError as exception:
        raise self.retry(exc=exception)
    else:
        event_file_metadata_document_file_finished.commit(
            action_object=document_file.document, actor=user,
            target=document_file
        )


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_metadata_driver_process(self, document_file_id, stored_driver_id):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )
    StoredDriver = apps.get_model(
        app_label='file_metadata', model_name='StoredDriver'
    )

    try:
        document_file = DocumentFile.objects.get(pk=document_file_id)

        stored_driver = StoredDriver.objects.get(pk=stored_driver_id)
    except OperationalError as exception:
        raise self.retry(exc=exception)
    else:
        lock_id = 'task_metadata_driver_process-{}-{}'.format(
            document_file_id, stored_driver_id
        )
        try:
            logger.debug('trying to acquire lock: %s', lock_id)
            lock = LockingBackend.get_backend().acquire_lock(
                name=lock_id, timeout=LOCK_EXPIRE
            )
            logger.debug('acquired lock: %s', lock_id)
        except LockError as exception:
            raise self.retry(exc=exception)
        except OperationalError as exception:
            raise self.retry(exc=exception)
        else:
            try:
                driver_class = stored_driver.driver_class

                driver_instance = driver_class()

                driver_instance.initialize()

                driver_instance.process(document_file=document_file)
            finally:
                lock.release()



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_document_download_multiple, icon_document_download_single,
    icon_document_file_download_quick
)
from .permissions import permission_document_file_download

link_document_download_multiple = Link(
    icon=icon_document_download_multiple, text=_(message='Download files'),
    view='document_downloads:document_download_multiple'
)
link_document_download_single = Link(
    args='resolved_object.pk', icon=icon_document_download_single,
    permission=permission_document_file_download,
    text=_(message='Download files'),
    view='document_downloads:document_download_single'
)
link_document_file_download_quick = Link(
    args='resolved_object.id', icon=icon_document_file_download_quick,
    permission=permission_document_file_download, tags='new_window',
    text=_(message='Quick download'),
    view='document_downloads:document_file_download'
)



from django.template import Library

from ..models import Announcement

register = Library()


@register.inclusion_tag('announcements/announcements.html')
def announcements():
    return {'announcements': Announcement.objects.get_for_now()}



from django.db import models

from mayan.apps.testing.tests.base import BaseTestCase

from ..search_models import SearchModel

from .mixins.base import SearchTestMixin


class SearchModelTestCase(SearchTestMixin, BaseTestCase):
    auto_test_search_objects_create = False

    def _create_test_models(self):
        self._TestModel = self._create_test_model(
            fields={
                'label': models.CharField(
                    max_length=32
                )
            }, model_name='TestModel'
        )

    def _setup_test_model_search(self):
        self._test_search_model = SearchModel(
            app_label=self._TestModel._meta.app_label,
            model_name=self._TestModel._meta.model_name
        )
        self._test_search_model.add_model_field(
            field='label'
        )

    def test_search_field_removal(self):
        test_search_fields = self._test_search_model.search_fields

        test_search_field = self._test_search_model.get_search_field(
            field_name='label'
        )

        self._test_search_model.remove_search_field(
            search_field=test_search_field
        )

        self.assertNotEqual(
            self._test_search_model.search_fields, test_search_fields
        )



import logging

from django.conf import settings
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.backends.views import (
    ViewSingleObjectDynamicFormModelBackendCreate,
    ViewSingleObjectDynamicFormModelBackendEdit
)
from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.views.generics import (
    ConfirmView, FormView, MultipleObjectConfirmActionView,
    SingleObjectDeleteView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..exceptions import SourceActionException
from ..forms import SourceBackendSelectionForm, SourceBackendSetupDynamicForm
from ..icons import (
    icon_source_action, icon_source_backend_selection, icon_source_create,
    icon_source_delete, icon_source_edit, icon_source_list, icon_source_test
)
from ..links import link_source_backend_selection
from ..models import Source
from ..permissions import (
    permission_sources_create, permission_sources_delete,
    permission_sources_edit, permission_sources_view
)
from ..source_backends.base import SourceBackend
from ..tasks import task_source_backend_action_execute

from .view_mixins import SourceActionViewMixin, SourceLinkNavigationViewMixin

logger = logging.getLogger(name=__name__)


class SourceActionView(
    SourceActionViewMixin, ExternalObjectViewMixin,
    SourceLinkNavigationViewMixin, MultipleObjectConfirmActionView
):
    external_object_pk_url_kwarg = 'source_id'
    external_object_queryset = Source.objects.filter(enabled=True)
    pk_url_kwarg = 'action_name'
    view_icon = icon_source_action

    def get_action_interface_kwargs(self):
        kwargs = {}
        kwargs.update(
            self.request.GET.dict()
        )
        kwargs.update(
            self.request.POST.dict()
        )
        kwargs['request'] = self.request
        return kwargs

    def get_context_data(self, **kwargs):
        context_super = super().get_context_data(**kwargs)

        try:
            context_action = self.object.get_confirmation_context(
                interface_name='View',
                interface_load_kwargs=self.get_action_interface_kwargs()
            )
        except SourceActionException as exception:
            messages.error(
                message=_(
                    'Unable to execute action; %s.'
                ) % exception, request=self.request
            )
            if settings.DEBUG:
                raise
        else:
            context_super.update(**context_action)

        return context_super

    def get_object_first(self):
        action = self.external_object.get_action(
            name=self.kwargs['action_name']
        )

        action_permission = action.permission

        if action_permission:
            source_queryset = self.external_object_queryset

            queryset = AccessControlList.objects.restrict_queryset(
                permission=action_permission, queryset=source_queryset,
                user=self.request.user
            )

            get_object_or_404(
                klass=queryset, pk=self.external_object.pk
            )

        return action

    def get_object_list(self):
        self.view_mode_single = True
        self.view_mode_multiple = False

        return list(
            self.external_object.get_action_list()
        )

    def get_source_link_action(self):
        return self.get_view_source_action()

    def get_source_link_permission(self):
        if 'document_id' in self.kwargs:
            return permission_document_file_new
        else:
            return permission_document_create

    def get_source_link_queryset(self):
        return self.queryset_source_valid

    def get_source_link_view_kwargs(self):
        if 'document_id' in self.kwargs:
            return {
                'document_id': self.kwargs['document_id']
            }
        else:
            return {}

    def get_source_link_view_name(self):
        if 'document_id' in self.kwargs:
            return 'sources:document_file_upload'
        else:
            return 'sources:document_upload'

    def get_view_source_action(self):
        if 'document_id' in self.kwargs:
            return 'document_file_upload'
        else:
            return 'document_upload'

    def view_action(self):
        try:
            result = self.object.execute(
                interface_name='View',
                interface_load_kwargs=self.get_action_interface_kwargs()
            )
        except Exception as exception:
            message = _(
                'Error processing source action; '
                '%(exception)s'
            ) % {
                'exception': exception,
            }
            logger.error(msg=message, exc_info=True)
            messages.error(
                message=message.replace('\n', ' '),
                request=self.request
            )
            if settings.DEBUG:
                raise
        else:
            messages.success(
                message=_(
                    'Source action executed successfully.'
                ), request=self.request
            )

            return result


class SourceBackendSelectionView(FormView):
    extra_context = {
        'title': _(message='New source backend selection')
    }
    form_class = SourceBackendSelectionForm
    view_icon = icon_source_backend_selection
    view_permission = permission_sources_create

    def form_valid(self, form):
        backend = form.cleaned_data['backend']
        return HttpResponseRedirect(
            redirect_to=reverse(
                kwargs={
                    'backend_path': backend
                }, viewname='sources:source_create'
            )
        )


class SourceCreateView(ViewSingleObjectDynamicFormModelBackendCreate):
    backend_class = SourceBackend
    form_class = SourceBackendSetupDynamicForm
    post_action_redirect = reverse_lazy(viewname='sources:source_list')
    view_icon = icon_source_create
    view_permission = permission_sources_create

    def get_extra_context(self):
        backend_class = self.get_backend_class()
        return {
            'title': _(
                'Create a "%s" source'
            ) % backend_class.label
        }

    def get_form_extra_kwargs(self):
        return {
            'user': self.request.user
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'backend_path': self.kwargs['backend_path']
        }


class SourceDeleteView(SingleObjectDeleteView):
    model = Source
    object_permission = permission_sources_delete
    pk_url_kwarg = 'source_id'
    post_action_redirect = reverse_lazy(
        viewname='sources:source_list'
    )
    view_icon = icon_source_delete

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Delete the source: %s?') % self.object
        }


class SourceEditView(ViewSingleObjectDynamicFormModelBackendEdit):
    form_class = SourceBackendSetupDynamicForm
    model = Source
    object_permission = permission_sources_edit
    pk_url_kwarg = 'source_id'
    post_action_redirect = reverse_lazy(
        viewname='sources:source_list'
    )
    view_icon = icon_source_edit

    def get_extra_context(self):
        return {
            'title': _(message='Edit source: %s') % self.object
        }

    def get_form_extra_kwargs(self):
        return {
            'user': self.request.user
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class SourceListView(SingleObjectListView):
    model = Source
    object_permission = permission_sources_view
    view_icon = icon_source_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_source_list,
            'no_results_main_link': link_source_backend_selection.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Sources provide the means to upload documents. '
                'Some sources are interactive and require user input to '
                'operate. Others are automatic and run in the background '
                'without user intervention.'
            ),
            'no_results_title': _(message='No sources available'),
            'title': _(message='Sources')
        }


class SourceTestView(ExternalObjectViewMixin, ConfirmView):
    """
    Trigger the task_source_backend_action_execute task for a given source to
    test/debug their configuration irrespective of the schedule task setup.
    """
    external_object_permission = permission_sources_edit
    external_object_pk_url_kwarg = 'source_id'
    external_object_class = Source
    view_icon = icon_source_test

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(request=request, *args, **kwargs)

        action_name = self.get_action_name()

        action = self.external_object.get_action(name=action_name)

        if not action.confirmation:
            messages.error(
                message=_(
                    'The selected action "%s" does not require '
                    'confirmation and cannot be tested.'
                ) % action_name, request=self.request
            )

            previous_url = self.get_previous_url()
            return HttpResponseRedirect(redirect_to=previous_url)
        else:
            return result

    def get_action_interface_kwargs(self):
        return {
            'dry_run': True
        }

    def get_action_name(self):
        return 'document_upload'

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'subtitle': _(
                'This will execute the source code even if the source '
                'is not enabled. Sources that delete content after '
                'downloading will not do so while being tested. Check the '
                'source\'s error log for information during testing. A '
                'successful test will clear the error log.'
            ), 'title': _(
                'Trigger check for source "%s"?'
            ) % self.external_object
        }

    def view_action(self):
        task_source_backend_action_execute.apply_async(
            kwargs={
                'action_interface_kwargs': self.get_action_interface_kwargs(),
                'action_name': self.get_action_name(),
                'source_id': self.external_object.id,
                'user_id': self.request.user.pk
            }
        )

        messages.success(
            message=_(
                'Source test queued. Check for newly created documents '
                'or for error log entries.'
            ), request=self.request
        )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_parsing', '0011_delete_documentfileparseerror')
    ]

    operations = [
        migrations.AlterField(
            model_name='documenttypesettings', name='auto_parsing',
            field=models.BooleanField(
                default=True, help_text='Automatically queue newly '
                'created documents for parsing.', verbose_name='Auto parsing'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

CHECK_EXPIRED_CHECK_OUTS_INTERVAL = 60  # Lowest check out expiration allowed
CHECKOUT_EXPIRATION_LOCK_EXPIRE = 50

STATE_CHECKED_OUT = 'checkedout'
STATE_CHECKED_IN = 'checkedin'

STATE_LABELS = {
    STATE_CHECKED_OUT: _(message='Checked out'),
    STATE_CHECKED_IN: _(message='Checked in/available'),
}



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_document_file_attachment_send_multiple,
    icon_document_file_attachment_send_single,
    icon_document_file_link_send_multiple,
    icon_document_file_link_send_single, icon_document_link_send_multiple,
    icon_document_link_send_single,
    icon_document_version_attachment_send_single,
    icon_document_version_attachment_send_multiple,
    icon_document_version_link_send_multiple,
    icon_document_version_link_send_single, icon_mailing_profile_create,
    icon_mailing_profile_delete, icon_mailing_profile_edit,
    icon_mailing_profile_list, icon_mailing_profile_test
)
from .permissions import (
    permission_mailing_profile_create, permission_mailing_profile_delete,
    permission_mailing_profile_edit, permission_mailing_profile_use,
    permission_mailing_profile_view,
    permission_send_document_file_attachment,
    permission_send_document_file_link, permission_send_document_link,
    permission_send_document_version_attachment,
    permission_send_document_version_link
)

# Document

link_send_document_link_single = Link(
    args='resolved_object.pk', icon=icon_document_link_send_single,
    permission=permission_send_document_link,
    text=_(message='Email document link'),
    view='mailer:send_document_link_single'
)
link_send_document_link_multiple = Link(
    icon=icon_document_link_send_multiple,
    text=_(message='Email document link'),
    view='mailer:send_document_link_multiple'
)

# Document file

link_send_document_file_attachment_single = Link(
    args='resolved_object.pk',
    icon=icon_document_file_attachment_send_single,
    permission=permission_send_document_file_attachment,
    text=_(message='Email document file'),
    view='mailer:send_document_file_attachment_single'
)
link_send_document_file_attachment_multiple = Link(
    icon=icon_document_file_attachment_send_multiple,
    text=_(message='Email document file'),
    view='mailer:send_document_file_attachment_multiple'
)
link_send_document_file_link_single = Link(
    args='resolved_object.pk', icon=icon_document_file_link_send_single,
    permission=permission_send_document_file_link,
    text=_(message='Email document file link'),
    view='mailer:send_document_file_link_single'
)
link_send_document_file_link_multiple = Link(
    icon=icon_document_file_link_send_multiple,
    text=_(message='Email document file link'),
    view='mailer:send_document_file_link_multiple'
)

# Document version

link_send_document_version_attachment_single = Link(
    args='resolved_object.pk',
    icon=icon_document_version_attachment_send_single,
    permission=permission_send_document_version_attachment,
    text=_(message='Email document version'),
    view='mailer:send_document_version_attachment_single'
)
link_send_document_version_attachment_multiple = Link(
    icon=icon_document_version_attachment_send_multiple,
    text=_(message='Email document version'),
    view='mailer:send_document_version_attachment_multiple'
)
link_send_document_version_link_single = Link(
    args='resolved_object.pk', icon=icon_document_version_link_send_single,
    permission=permission_send_document_version_link,
    text=_(message='Email document version link'),
    view='mailer:send_document_version_link_single'
)
link_send_document_version_link_multiple = Link(
    icon=icon_document_version_link_send_multiple,
    text=_(message='Email link version'),
    view='mailer:send_document_version_link_multiple'
)

# Mailing profile

link_mailing_profile_create = Link(
    icon=icon_mailing_profile_create,
    permission=permission_mailing_profile_create,
    text=_(message='Create mailing profile'),
    view='mailer:mailing_profile_backend_selection'
)
link_mailing_profile_delete = Link(
    args='resolved_object.pk', icon=icon_mailing_profile_delete,
    permission=permission_mailing_profile_delete, tags='dangerous',
    text=_(message='Delete'), view='mailer:mailing_profile_delete'
)
link_mailing_profile_edit = Link(
    args='object.pk', icon=icon_mailing_profile_edit,
    permission=permission_mailing_profile_edit, text=_(message='Edit'),
    view='mailer:mailing_profile_edit'
)
link_mailing_profile_list = Link(
    icon=icon_mailing_profile_list, text=_(message='Mailing profiles'),
    view='mailer:mailing_profile_list'
)
link_mailing_profile_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='mailer', model_name='UserMailer',
        object_permission=permission_mailing_profile_view,
        view_permission=permission_mailing_profile_create,
    ), icon=icon_mailing_profile_list, text=_(message='Mailing profiles'),
    view='mailer:mailing_profile_list'
)
link_mailing_profile_list = Link(
    icon=icon_mailing_profile_list,
    permission=permission_mailing_profile_view,
    text=_(message='Mailing profiles'), view='mailer:mailing_profile_list'
)
link_mailing_profile_test = Link(
    args='object.pk', icon=icon_mailing_profile_test,
    permission=permission_mailing_profile_use, text=_(message='Test'),
    view='mailer:mailing_profile_test'
)



from mayan.apps.rest_api import generics

from ..models.document_models import RecentlyCreatedDocument
from ..permissions import permission_document_view
from ..serializers.document_serializers import DocumentSerializer


class APIRecentlyCreatedDocumentListView(generics.ListAPIView):
    """
    get: Return a list of the recently created documents.
    """
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = DocumentSerializer
    source_queryset = RecentlyCreatedDocument.valid.all()



from .base import SearchTestMixin


class SearchToolsViewTestMixin:
    def _request_search_backend_reindex_view(self):
        return self.post(viewname='search:search_backend_reindex')


class SearchViewTestMixin(SearchTestMixin):
    def _request_search_advanced_get_view(self):
        return self.get(
            viewname='search:search_advanced', kwargs={
                'search_model_pk': self._test_search_model.full_name
            }
        )

    def _request_search_again_view(self, follow=False, query=None):
        return self.post(
            follow=follow, viewname='search:search_again', kwargs={
                'search_model_pk': self._test_search_model.full_name
            }, query=query
        )

    def _request_search_simple_get_view(self):
        return self.get(
            viewname='search:search_simple', kwargs={
                'search_model_pk': self._test_search_model.full_name
            }
        )

    def _request_search_results_view(self, data, kwargs=None, query=None):
        return self.get(
            viewname='search:search_results', data=data, kwargs=kwargs,
            query=query
        )



from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_HYBRID_PDF_CONTENT

from .mixins import DocumentFileContentTestMixin


class DocumentFileContentMethodTestCase(
    DocumentFileContentTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def test_document_file_parsed_content(self):
        self._upload_test_document()

        self._create_test_document_file_parsed_content()
        value = ' '.join(
            self._test_document_file.content()
        )

        self.assertEqual(value, TEST_FILE_HYBRID_PDF_CONTENT)

    def test_document_parsed_content(self):
        self._upload_test_document()
        self._create_test_document_file_parsed_content()
        value = ' '.join(
            self._test_document.content()
        )

        self.assertEqual(value, TEST_FILE_HYBRID_PDF_CONTENT)

    def test_document_stub_parsed_content(self):
        self._create_test_document_stub()

        value = ' '.join(
            self._test_document.content()
        )

        self.assertFalse(value)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0050_auto_20190725_0451')
    ]

    operations = [
        migrations.AddField(
            model_name='documentpage',
            name='enabled',
            field=models.BooleanField(default=True, verbose_name='Enabled')
        )
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_comments', '0005_auto_20210130_0658')
    ]

    operations = [
        migrations.AlterField(
            model_name='comment',
            name='text',
            field=models.TextField(verbose_name='Text'),
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.storage.classes import DefinedStorage

from .literals import STORAGE_NAME_SOURCE_CACHE_FOLDER
from .settings import (
    setting_source_cache_storage_backend,
    setting_source_cache_storage_backend_arguments
)

storage_source_cache = DefinedStorage(
    dotted_path=setting_source_cache_storage_backend.value,
    error_message=_(
        'Unable to initialize the staging folder file image '
        'storage. Check the settings {} and {} for formatting '
        'errors.'.format(
            setting_source_cache_storage_backend.global_name,
            setting_source_cache_storage_backend_arguments.global_name
        )
    ),
    label=_(message='Staging folder files'),
    name=STORAGE_NAME_SOURCE_CACHE_FOLDER,
    kwargs=setting_source_cache_storage_backend_arguments.value
)



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..events import event_tag_attached, event_tag_removed

from .mixins import TagTestMixin


class TagDocumentTestCase(DocumentTestMixin, TagTestMixin, BaseTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_method_document_attach(self):
        self._create_test_tag()

        self._clear_events()

        self._test_tag.attach_to(
            document=self._test_document, user=self._test_case_user
        )

        self.assertTrue(
            self._test_document in self._test_tag.documents.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_attached.id)

    def test_method_document_count(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        self.assertEqual(
            self._test_tag.get_document_count(user=self._test_case_user),
            len(self._test_documents)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_method_document_count_with_trashed_document(self):
        self._create_test_tag(add_test_document=True)
        self._test_document.delete()

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        self.assertEqual(
            self._test_tag.get_document_count(user=self._test_case_user),
            len(self._test_documents) - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_method_document_remove(self):
        self._create_test_tag(add_test_document=True)

        self._clear_events()

        self._test_tag.remove_from(
            document=self._test_document, user=self._test_case_user
        )

        self.assertTrue(
            self._test_document not in self._test_tag.documents.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_removed.id)

    def test_method_get_absolute_url(self):
        self._create_test_tag()

        self._clear_events()

        self.assertTrue(
            self._test_tag.get_absolute_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.db import models, transaction
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..events import (
    event_document_version_created, event_document_version_deleted,
    event_document_version_edited
)
from ..managers import ValidDocumentVersionManager

from .document_models import Document
from .document_version_model_mixins import DocumentVersionBusinessLogicMixin

__all__ = ('DocumentVersion', 'DocumentVersionSearchResult')


class DocumentVersion(
    DocumentVersionBusinessLogicMixin, ExtraDataModelMixin, models.Model
):
    document = models.ForeignKey(
        on_delete=models.CASCADE, related_name='versions', to=Document,
        verbose_name=_(message='Document')
    )
    timestamp = models.DateTimeField(
        auto_now_add=True, db_index=True, help_text=_(
            'The server date and time when the document version was created.'
        ), verbose_name=_(message='Timestamp')
    )
    comment = models.TextField(
        blank=True, default='', help_text=_(
            'An optional short text describing the document version.'
        ), verbose_name=_(message='Comment')
    )
    active = models.BooleanField(
        default=False, help_text=_(
            'Determines the active version of the document.'
        ), verbose_name=_(message='Active')
    )

    objects = models.Manager()
    valid = ValidDocumentVersionManager()

    class Meta:
        ordering = ('timestamp',)
        verbose_name = _(message='Document version')
        verbose_name_plural = _(message='Document versions')

    def __str__(self):
        return self.get_label()

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_document_version_deleted,
        target='document'
    )
    def delete(self, *args, **kwargs):
        for page in self.pages.all():
            page.delete()

        self.cache_partition.delete()

        return super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse(
            viewname='documents:document_version_preview', kwargs={
                'document_version_id': self.pk
            }
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_document_version_created,
            'action_object': 'document',
            'target': 'self'
        },
        edited={
            'event': event_document_version_edited,
            'action_object': 'document',
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        new_document_version = not self.pk

        if new_document_version:
            with transaction.atomic():
                result = super().save(*args, **kwargs)

                if self.active:
                    # Don't save the same version again. The active value
                    # was already save in the previous super call to
                    # .save().
                    self.active_set(save=False)

                return result
        else:
            # Handle existing document versions that change the value of
            # active directly without using the method `active_set`. Such
            # as via the API or direct model manipulation.
            if self.active:
                # Don't save the version. The active value will be made
                # permanent in the `super.save` call following this
                # statement.
                self.active_set(save=False)

            return super().save(*args, **kwargs)


class DocumentVersionSearchResult(DocumentVersion):
    class Meta:
        proxy = True



from django.db import models

from ...search_query_types import (
    QueryTypeExact, QueryTypeFuzzy, QueryTypeGreaterThan,
    QueryTypeGreaterThanOrEqual, QueryTypeLessThan, QueryTypeLessThanOrEqual,
    QueryTypePartial, QueryTypeRange, QueryTypeRangeExclusive,
    QueryTypeRegularExpression
)
from ...value_transformations import (
    ValueTransformationHyphenReplace, ValueTransformationHyphenStrip,
    ValueTransformationToBoolean, ValueTransformationToDateTime,
    ValueTransformationToInteger, ValueTransformationToString
)

DEFAULT_FUZZY_SLOP = 2

DJANGO_TO_DJANGO_FIELD_MAP = {
    models.AutoField: {
        'field': models.AutoField,
        'query_type_list': [
            QueryTypeExact, QueryTypePartial, QueryTypeGreaterThan,
            QueryTypeGreaterThanOrEqual, QueryTypeLessThan,
            QueryTypeLessThanOrEqual, QueryTypeRange, QueryTypeRangeExclusive
        ],
        'transformations': {
            'search': [ValueTransformationToInteger]
        }
    },
    models.BooleanField: {
        'field': models.BooleanField,
        'query_type_list': [
            QueryTypeExact
        ],
        'transformations': {
            'search': [
                ValueTransformationToString, ValueTransformationToBoolean
            ]
        }
    },
    models.CharField: {
        'field': models.CharField,
        'query_type_list': [
            QueryTypeExact, QueryTypeFuzzy, QueryTypePartial,
            QueryTypeRegularExpression, QueryTypeGreaterThan,
            QueryTypeGreaterThanOrEqual, QueryTypeLessThan,
            QueryTypeLessThanOrEqual
        ],
        'transformations': {
            'search': [
                ValueTransformationToString, ValueTransformationHyphenReplace
            ]
        }
    },
    models.DateTimeField: {
        'field': models.DateTimeField,
        'query_type_list': [
            QueryTypeExact, QueryTypeGreaterThan,
            QueryTypeGreaterThanOrEqual, QueryTypeLessThan,
            QueryTypeLessThanOrEqual, QueryTypeRange, QueryTypeRangeExclusive
        ],
        'transformations': {
            'search': [ValueTransformationToDateTime]
        }
    },
    models.EmailField: {
        'field': models.EmailField,
        'query_type_list': [
            QueryTypeExact, QueryTypePartial
        ],
        'transformations': {
            'search': [ValueTransformationToString]
        }
    },
    models.IntegerField: {
        'field': models.IntegerField,
        'query_type_list': [
            QueryTypeExact, QueryTypeGreaterThan,
            QueryTypeGreaterThanOrEqual, QueryTypeLessThan,
            QueryTypeLessThanOrEqual, QueryTypeRange, QueryTypeRangeExclusive
        ],
        'transformations': {
            'search': [ValueTransformationToInteger]
        }
    },
    models.PositiveIntegerField: {
        'field': models.PositiveIntegerField,
        'query_type_list': [
            QueryTypeExact, QueryTypeGreaterThan,
            QueryTypeGreaterThanOrEqual, QueryTypeLessThan,
            QueryTypeLessThanOrEqual, QueryTypeRange, QueryTypeRangeExclusive
        ],
        'transformations': {
            'search': [ValueTransformationToInteger]
        }
    },
    models.TextField: {
        'field': models.TextField,
        'query_type_list': [
            QueryTypeExact, QueryTypeFuzzy, QueryTypePartial,
            QueryTypeRegularExpression, QueryTypeGreaterThan,
            QueryTypeGreaterThanOrEqual, QueryTypeLessThan,
            QueryTypeLessThanOrEqual
        ],
        'transformations': {
            'search': [
                ValueTransformationToString, ValueTransformationHyphenReplace
            ]
        }
    },
    models.UUIDField: {
        'field': models.UUIDField,
        'query_type_list': [
            QueryTypeExact, QueryTypePartial, QueryTypeRegularExpression
        ],
        'transformations': {
            'search': [
                ValueTransformationToString, ValueTransformationHyphenStrip
            ]
        }
    }
}

MAXIMUM_FUZZY_OPTIONS = 50



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Sources'), name='sources'
)

event_source_created = namespace.add_event_type(
    label=_(message='Source created'), name='source_create'
)
event_source_edited = namespace.add_event_type(
    label=_(message='Source edited'), name='source_edit'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_document_type_web_links, icon_document_web_link_list,
    icon_web_link_create, icon_web_link_delete, icon_web_link_document_types,
    icon_web_link_edit, icon_web_link_instance_view, icon_web_link_list,
    icon_web_link_setup
)
from .permissions import (
    permission_web_link_create, permission_web_link_delete,
    permission_web_link_edit, permission_web_link_instance_view,
    permission_web_link_view
)

link_document_type_web_links = Link(
    args='resolved_object.pk', icon=icon_document_type_web_links,
    permission=permission_document_type_edit, text=_(message='Web links'),
    view='web_links:document_type_web_links'
)
link_document_web_link_list = Link(
    args='resolved_object.pk', icon=icon_document_web_link_list,
    permission=permission_web_link_instance_view, text=_(message='Web links'),
    view='web_links:document_web_link_list'
)
link_web_link_create = Link(
    icon=icon_web_link_create,
    permission=permission_web_link_create, text=_(message='Create new web link'),
    view='web_links:web_link_create'
)
link_web_link_delete = Link(
    args='object.pk', icon=icon_web_link_delete,
    permission=permission_web_link_delete,
    tags='dangerous', text=_(message='Delete'), view='web_links:web_link_delete'
)
link_web_link_document_types = Link(
    args='object.pk', icon=icon_web_link_document_types,
    permission=permission_web_link_edit, text=_(message='Document types'),
    view='web_links:web_link_document_types'
)
link_web_link_edit = Link(
    args='object.pk', icon=icon_web_link_edit,
    permission=permission_web_link_edit, text=_(message='Edit'),
    view='web_links:web_link_edit'
)
link_web_link_instance_view = Link(
    icon=icon_web_link_instance_view,
    args=('document.pk', 'object.pk',),
    permission=permission_web_link_instance_view, tags='new_window',
    text=_(message='Navigate'), view='web_links:web_link_instance_view'
)
link_web_link_list = Link(
    icon=icon_web_link_list, text=_(message='Web links'),
    view='web_links:web_link_list'
)
link_web_link_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='web_links', model_name='WebLink',
        object_permission=permission_web_link_view,
        view_permission=permission_web_link_create,
    ), icon=icon_web_link_setup, text=_(message='Web links'),
    view='web_links:web_link_list'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_PLATFORM_CLIENT_BACKEND_ARGUMENTS,
    DEFAULT_PLATFORM_CLIENT_BACKEND_ENABLED
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Platform'), name='platform'
)

setting_client_backend_enabled = setting_namespace.do_setting_add(
    default=DEFAULT_PLATFORM_CLIENT_BACKEND_ENABLED,
    global_name='PLATFORM_CLIENT_BACKEND_ENABLED', help_text=_(
        'List of client backends to launch after startup. Use full dotted '
        'path to the client backend classes.'
    )
)
setting_client_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_PLATFORM_CLIENT_BACKEND_ARGUMENTS,
    global_name='PLATFORM_CLIENT_BACKEND_ARGUMENTS', help_text=_(
        'Arguments for the client backends. Use the client backend dotted '
        'path as the dictionary key for the arguments in dictionary format.'
    )
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0008_auto_20150711_0723'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transformation',
            name='order',
            field=models.PositiveIntegerField(
                default=0, help_text='Order in which the transformations '
                'will be executed. If left unchanged, an automatic order '
                'value will be assigned.', db_index=True,
                verbose_name='Order', blank=True
            ),
            preserve_default=True,
        ),
    ]



from django.core.management.base import BaseCommand

from ...classes import Statistic


class Command(BaseCommand):
    help = 'Remove obsolete statistics scheduled and results from the database'

    def handle(self, *args, **options):
        Statistic.purge_schedules()



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Web links'), name='web_links'
)

permission_web_link_create = namespace.add_permission(
    label=_(message='Create new web links'), name='web_link_create'
)
permission_web_link_delete = namespace.add_permission(
    label=_(message='Delete web links'), name='web_link_delete'
)
permission_web_link_edit = namespace.add_permission(
    label=_(message='Edit web links'), name='web_link_edit'
)
permission_web_link_view = namespace.add_permission(
    label=_(message='View existing web links'), name='web_link_view'
)
permission_web_link_instance_view = namespace.add_permission(
    label=_(message='View web link instances'), name='web_link_instance_view'
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('mailer', '0005_auto_20170718_0123')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='usermailer',
            options={
                'ordering': ('label',),
                'verbose_name': 'Mailing profile',
                'verbose_name_plural': 'Mailing profiles'
            }
        ),
        migrations.AlterField(
            model_name='usermailer',
            name='label',
            field=models.CharField(
                help_text='A short text describing the mailing profile.',
                max_length=128, unique=True, verbose_name='Label'
            )
        )
    ]



from django.core import validators
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ValueChangeModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerSave
from mayan.apps.lock_manager.decorators import locked_class_method

from .events import event_cache_created, event_cache_edited
from .model_mixins import (
    CacheBusinessLogicMixin, CachePartitionBusinessLogicMixin,
    CachePartitionFileBusinessLogicMixin
)


class Cache(CacheBusinessLogicMixin, ValueChangeModelMixin, models.Model):
    defined_storage_name = models.CharField(
        db_index=True, help_text=_(
            'Internal name of the defined storage for this cache.'
        ), max_length=96, unique=True, verbose_name=_(
            message='Defined storage name'
        )
    )
    maximum_size = models.PositiveBigIntegerField(
        db_index=True, help_text=_(
            message='Maximum size of the cache in bytes.'
        ), validators=[
            validators.MinValueValidator(limit_value=1)
        ], verbose_name=_(message='Maximum size')
    )

    class Meta:
        ordering = ('id',)
        verbose_name = _(message='Cache')
        verbose_name_plural = _(message='Caches')

    def __str__(self):
        return str(self.label)

    def get_absolute_url(self):
        return reverse(
            viewname='file_caching:cache_detail', kwargs={
                'cache_id': self.pk
            }
        )

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_cache_created,
            'target': 'self'
        },
        edited={
            'event': event_cache_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        old_maximum_size = self._get_field_previous_value(
            field='maximum_size'
        )

        result = super().save(*args, **kwargs)

        if self.maximum_size < old_maximum_size:
            self.prune()

        return result


class CachePartition(CachePartitionBusinessLogicMixin, models.Model):
    cache = models.ForeignKey(
        on_delete=models.CASCADE, related_name='partitions',
        to=Cache, verbose_name=_(message='Cache')
    )
    name = models.CharField(
        max_length=128, verbose_name=_(message='Name')
    )

    class Meta:
        unique_together = ('cache', 'name')
        verbose_name = _(message='Cache partition')
        verbose_name_plural = _(message='Cache partitions')

    def __str__(self):
        return '{} ({})'.format(self.cache, self.name)

    def delete(self, *args, **kwargs):
        self.purge(user=self)
        return super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse(
            viewname='file_caching:cache_partition_detail', kwargs={
                'cache_partition_id': self.pk
            }
        )


class CachePartitionFile(CachePartitionFileBusinessLogicMixin, models.Model):
    _storage_object = None

    partition = models.ForeignKey(
        on_delete=models.CASCADE, related_name='files',
        to=CachePartition, verbose_name=_(message='Cache partition')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_(message='Date time')
    )
    filename = models.CharField(
        max_length=255, verbose_name=_(message='Filename')
    )
    file_size = models.PositiveIntegerField(
        default=0, verbose_name=_(message='File size')
    )
    hits = models.PositiveIntegerField(
        db_index=True, default=0, help_text=_(
            'Times this cache partition file has been accessed.'
        ), verbose_name='Hits'
    )

    class Meta:
        get_latest_by = 'datetime'
        unique_together = ('partition', 'filename')
        verbose_name = _(message='Cache partition file')
        verbose_name_plural = _(message='Cache partition files')

    @locked_class_method
    def delete(self, *args, **kwargs):
        storage_instance = self.partition.cache.storage
        storage_instance.delete(name=self.full_filename)
        return super().delete(*args, **kwargs)



from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..events import event_workflow_template_edited

from .workflow_models import Workflow
from .workflow_state_model_mixins import WorkflowStateBusinessLogicMixin

__all__ = ('WorkflowState', 'WorkflowStateRuntimeProxy')


class WorkflowState(
    ExtraDataModelMixin, WorkflowStateBusinessLogicMixin, models.Model
):
    """
    Fields:
    * completion - Completion Amount - A user defined numerical value to help
    determine if the workflow of the document is nearing completion (100%).
    The Completion Amount will be determined by the completion value of the
    Actual State. Example: If the workflow has 3 states: registered, approved,
    archived; the admin could give the follow completion values to the
    states: 33%, 66%, 100%. If the Actual State of the document if approved,
    the Completion Amount will show 66%.
    """
    workflow = models.ForeignKey(
        on_delete=models.CASCADE, related_name='states', to=Workflow,
        verbose_name=_(message='Workflow')
    )
    label = models.CharField(
        help_text=_(message='A short text to describe the workflow state.'),
        max_length=255, verbose_name=_(message='Label')
    )
    initial = models.BooleanField(
        default=False,
        help_text=_(
            'The state at which the workflow will start in. Only one state '
            'can be the initial state.'
        ), verbose_name=_(message='Initial')
    )
    completion = models.IntegerField(
        blank=True, default=0, help_text=_(
            'The percent of completion that this state represents in '
            'relation to the workflow. Use numbers without the percent sign.'
        ), verbose_name=_(message='Completion')
    )

    class Meta:
        ordering = ('label',)
        unique_together = ('workflow', 'label')
        verbose_name = _(message='Workflow state')
        verbose_name_plural = _(message='Workflow states')

    def __str__(self):
        return self.label

    @method_event(
        action_object='self',
        event_manager_class=EventManagerMethodAfter,
        event=event_workflow_template_edited,
        target='workflow'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'workflow'
        },
        edited={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'workflow'
        }
    )
    def save(self, *args, **kwargs):
        # Solve issue #557 "Break workflows with invalid input"
        # without using a migration.
        # Remove blank=True, remove this, and create a migration in the next
        # minor version.

        try:
            self.completion = int(self.completion)
        except (TypeError, ValueError):
            self.completion = 0

        if self.initial:
            self.workflow.states.all().update(initial=False)
        return super().save(*args, **kwargs)


class WorkflowStateRuntimeProxy(WorkflowState):
    class Meta:
        proxy = True
        verbose_name = _(message='Workflow state runtime proxy')
        verbose_name_plural = _(message='Workflow state runtime proxies')

    def get_documents(self, permission=None, user=None):
        """
        Provide a queryset of the documents. The queryset is optionally
        filtered by access.
        """
        queryset = super().get_documents()

        if permission and user:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=user
            )

        return queryset

    def get_document_count(self, user):
        """
        Return the numeric count of documents at this workflow state.
        The count is filtered by access.
        """
        return self.get_documents(
            permission=permission_document_view, user=user
        ).count()



from django import forms

from .field_mixins import (
    FormFieldMixinFilteredQueryset, ModelFieldMixinFilteredQuerySet
)


class FormFieldFilteredModelChoice(
    FormFieldMixinFilteredQueryset, forms.ChoiceField
):
    """Single selection filtered model choice field."""


class FormFieldFilteredModelChoiceMultiple(
    FormFieldMixinFilteredQueryset, forms.MultipleChoiceField
):
    """Multiole selection filtered model choice field."""


class ModelFormFieldFilteredModelChoice(
    ModelFieldMixinFilteredQuerySet, forms.ModelChoiceField
):
    """Single selection filtered model choice field."""


class ModelFormFieldFilteredModelMultipleChoice(
    ModelFieldMixinFilteredQuerySet, forms.ModelMultipleChoiceField
):
    """Multiple selection filtered model choice field."""



from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='UpdatedSetting',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                (
                    'global_name', models.CharField(
                        db_index=True, help_text='A short text used as '
                        'the tag name.', max_length=255, unique=True,
                        verbose_name='Global name'
                    )
                ),
                (
                    'value_new', models.TextField(
                        blank=True, null=True, verbose_name='New value'
                    )
                ),
                (
                    'value_old', models.TextField(
                        blank=True, null=True, verbose_name='Old value'
                    )
                )
            ],
            options={
                'verbose_name': 'Updated setting',
                'verbose_name_plural': 'Updated settings'
            }
        )
    ]



from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_edited, event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .literals import (
    TEST_EMAIL_ATTACHMENT_AND_INLINE, TEST_EMAIL_BASE64_FILENAME,
    TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME, TEST_EMAIL_INLINE_IMAGE,
    TEST_EMAIL_NO_CONTENT_TYPE, TEST_EMAIL_NO_CONTENT_TYPE_STRING,
    TEST_EMAIL_ZERO_LENGTH_ATTACHMENT
)
from .mixins import (
    EmailSourceTestMixin, IMAPEmailSourceTestMixin, POP3EmailSourceTestMixin
)


class EmailSourceBackendActionDocumentUploadTestCase(
    EmailSourceTestMixin, GenericDocumentTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def test_decode_email_base64_encoded_filename(self):
        """
        Test decoding of base64 encoded e-mail attachment filename.
        """
        self._test_source_content = TEST_EMAIL_BASE64_FILENAME
        self._test_source_create()

        test_document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )

        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_decode_email_no_content_type(self):
        self._test_source_content = TEST_EMAIL_NO_CONTENT_TYPE
        self._test_source_create(
            extra_data={
                'store_body': True
            }
        )

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertTrue(
            TEST_EMAIL_NO_CONTENT_TYPE_STRING in Document.objects.first().file_latest.open().read()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_decode_email_zero_length_attachment(self):
        self._test_source_content = TEST_EMAIL_ZERO_LENGTH_ATTACHMENT
        self._test_source_create()

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(
            Document.objects.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_decode_email_with_inline_image(self):
        self._test_source_content = TEST_EMAIL_INLINE_IMAGE
        self._test_source_create(
            extra_data={
                'store_body': True
            }
        )

        self._silence_logger(name='mayan.apps.converter.backends')

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(
            Document.objects.count(), 2
        )
        self.assertQuerySetEqual(
            ordered=False, qs=Document.objects.all(), transform=repr, values=(
                '<Document: test-01.png>', '<Document: email_body.html>'
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 10)

        test_documents = Document.objects.all()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_documents[0])
        self.assertEqual(events[0].target, test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_documents[0])
        self.assertEqual(events[1].actor, test_documents[0].file_latest)
        self.assertEqual(events[1].target, test_documents[0].file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_documents[0])
        self.assertEqual(events[2].actor, test_documents[0].file_latest)
        self.assertEqual(events[2].target, test_documents[0].file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_documents[0])
        self.assertEqual(events[3].actor, test_documents[0].version_active)
        self.assertEqual(events[3].target, test_documents[0].version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, test_documents[0].version_active
        )
        self.assertEqual(
            events[4].actor, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].target, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_documents[0])
        self.assertEqual(events[5].actor, test_documents[0].version_active)
        self.assertEqual(events[5].target, test_documents[0].version_active)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, self._test_document_type)
        self.assertEqual(events[6].actor, test_documents[1])
        self.assertEqual(events[6].target, test_documents[1])
        self.assertEqual(events[6].verb, event_document_created.id)

        self.assertEqual(events[7].action_object, test_documents[1])
        self.assertEqual(events[7].actor, test_documents[1].file_latest)
        self.assertEqual(events[7].target, test_documents[1].file_latest)
        self.assertEqual(events[7].verb, event_document_file_created.id)

        self.assertEqual(events[8].action_object, test_documents[1])
        self.assertEqual(events[8].actor, test_documents[1].version_active)
        self.assertEqual(events[8].target, test_documents[1].version_active)
        self.assertEqual(events[8].verb, event_document_version_created.id)

        self.assertEqual(events[9].action_object, test_documents[1])
        self.assertEqual(events[9].actor, test_documents[1].version_active)
        self.assertEqual(events[9].target, test_documents[1].version_active)
        self.assertEqual(events[9].verb, event_document_version_edited.id)

    def test_decode_email_with_attachment_and_inline_image(self):
        self._test_source_content = TEST_EMAIL_ATTACHMENT_AND_INLINE
        self._test_source_create(
            extra_data={
                'store_body': True
            }
        )

        # Silence expected errors in other apps.
        self._silence_logger(name='mayan.apps.converter.backends')

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(
            Document.objects.count(), 2
        )
        self.assertQuerySetEqual(
            ordered=False, qs=Document.objects.all(), transform=repr, values=(
                '<Document: test-01.png>', '<Document: email_body.html>',
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 10)

        test_documents = Document.objects.all()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_documents[0])
        self.assertEqual(events[0].target, test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_documents[0])
        self.assertEqual(events[1].actor, test_documents[0].file_latest)
        self.assertEqual(events[1].target, test_documents[0].file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_documents[0])
        self.assertEqual(events[2].actor, test_documents[0].file_latest)
        self.assertEqual(events[2].target, test_documents[0].file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_documents[0])
        self.assertEqual(events[3].actor, test_documents[0].version_active)
        self.assertEqual(events[3].target, test_documents[0].version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, test_documents[0].version_active
        )
        self.assertEqual(
            events[4].actor, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].target, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_documents[0])
        self.assertEqual(events[5].actor, test_documents[0].version_active)
        self.assertEqual(events[5].target, test_documents[0].version_active)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, self._test_document_type)
        self.assertEqual(events[6].actor, test_documents[1])
        self.assertEqual(events[6].target, test_documents[1])
        self.assertEqual(events[6].verb, event_document_created.id)

        self.assertEqual(events[7].action_object, test_documents[1])
        self.assertEqual(events[7].actor, test_documents[1].file_latest)
        self.assertEqual(events[7].target, test_documents[1].file_latest)
        self.assertEqual(events[7].verb, event_document_file_created.id)

        self.assertEqual(events[8].action_object, test_documents[1])
        self.assertEqual(events[8].actor, test_documents[1].version_active)
        self.assertEqual(events[8].target, test_documents[1].version_active)
        self.assertEqual(events[8].verb, event_document_version_created.id)

        self.assertEqual(events[9].action_object, test_documents[1])
        self.assertEqual(events[9].actor, test_documents[1].version_active)
        self.assertEqual(events[9].target, test_documents[1].version_active)
        self.assertEqual(events[9].verb, event_document_version_edited.id)

    def test_document_upload_no_body(self):
        self._test_source_content = TEST_EMAIL_ATTACHMENT_AND_INLINE
        self._test_source_create()
        # Silence expected errors in other apps.
        self._silence_logger(name='mayan.apps.converter.backends')

        test_document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        # Only two attachments, no body document.
        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_edited.id)

    def test_document_upload_with_body(self):
        self._test_source_content = TEST_EMAIL_ATTACHMENT_AND_INLINE
        self._test_source_create(
            extra_data={
                'store_body': True
            }
        )

        # Silence expected errors in other apps.
        self._silence_logger(name='mayan.apps.converter.backends')

        test_document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(
            Document.objects.count(), test_document_count + 2
        )

        # Only two attachments and a body document.
        self.assertEqual(
            Document.objects.count(), 2
        )
        self.assertEqual(
            Document.objects.first().label, 'email_body.html'
        )
        self.assertEqual(
            Document.objects.all()[0].label, 'email_body.html'
        )
        self.assertEqual(
            Document.objects.all()[1].label, 'test-01.png'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 10)

        test_documents = Document.objects.all()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_documents[0])
        self.assertEqual(events[0].target, test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_documents[0])
        self.assertEqual(events[1].actor, test_documents[0].file_latest)
        self.assertEqual(events[1].target, test_documents[0].file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_documents[0])
        self.assertEqual(events[2].actor, test_documents[0].file_latest)
        self.assertEqual(events[2].target, test_documents[0].file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_documents[0])
        self.assertEqual(events[3].actor, test_documents[0].version_active)
        self.assertEqual(events[3].target, test_documents[0].version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, test_documents[0].version_active
        )
        self.assertEqual(
            events[4].actor, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].target, test_documents[0].version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_documents[0])
        self.assertEqual(events[5].actor, test_documents[0].version_active)
        self.assertEqual(events[5].target, test_documents[0].version_active)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

        self.assertEqual(events[6].action_object, self._test_document_type)
        self.assertEqual(events[6].actor, test_documents[1])
        self.assertEqual(events[6].target, test_documents[1])
        self.assertEqual(events[6].verb, event_document_created.id)

        self.assertEqual(events[7].action_object, test_documents[1])
        self.assertEqual(events[7].actor, test_documents[1].file_latest)
        self.assertEqual(events[7].target, test_documents[1].file_latest)
        self.assertEqual(events[7].verb, event_document_file_created.id)

        self.assertEqual(events[8].action_object, test_documents[1])
        self.assertEqual(events[8].actor, test_documents[1].version_active)
        self.assertEqual(events[8].target, test_documents[1].version_active)
        self.assertEqual(events[8].verb, event_document_version_created.id)

        self.assertEqual(events[9].action_object, test_documents[1])
        self.assertEqual(events[9].actor, test_documents[1].version_active)
        self.assertEqual(events[9].target, test_documents[1].version_active)
        self.assertEqual(events[9].verb, event_document_version_edited.id)


class IMAPSourceBackendActionDocumentUploadTestCase(
    IMAPEmailSourceTestMixin, GenericDocumentTestCase
):
    _test_source_content = TEST_EMAIL_BASE64_FILENAME
    auto_upload_test_document = False

    def test_upload_simple_file(self):
        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_dry_run_false(self):
        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        self._execute_test_source_action(
            action_name='document_upload', extra_data={'dry_run': False}
        )

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_dry_run_none(self):
        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        self._execute_test_source_action(
            action_name='document_upload', extra_data={'dry_run': None}
        )

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_dry_run_true(self):
        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        self._execute_test_source_action(
            action_name='document_upload', extra_data={'dry_run': True}
        )

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)


class POP3SourceBackendActionDocumentUploadTestCase(
    POP3EmailSourceTestMixin, GenericDocumentTestCase
):
    _test_source_content = TEST_EMAIL_BASE64_FILENAME
    auto_upload_test_document = False

    def test_upload_simple_file(self):
        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_dry_run_false(self):
        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        self._execute_test_source_action(
            action_name='document_upload', extra_data={'dry_run': False}
        )

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_dry_run_none(self):
        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        self._execute_test_source_action(
            action_name='document_upload', extra_data={'dry_run': None}
        )

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)

    def test_dry_run_true(self):
        test_document_count = Document.objects.count()

        test_source_stored_file_count = len(
            self.get_test_source_stored_file_list()
        )

        self._clear_events()

        self._execute_test_source_action(
            action_name='document_upload', extra_data={'dry_run': True}
        )

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

        self.assertEqual(
            len(
                self.get_test_source_stored_file_list()
            ), test_source_stored_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, test_document_version_page)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, test_document)
        self.assertEqual(events[5].actor, test_document_version)
        self.assertEqual(events[5].target, test_document_version)
        self.assertEqual(events[5].verb, event_document_version_edited.id)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0013_auto_20170714_2133')
    ]

    operations = [
        migrations.AlterField(
            model_name='indextemplatenode',
            name='expression',
            field=models.TextField(
                help_text="Enter a template to render. Use Django's default "
                "templating language "
                "(https://docs.djangoproject.com/en/1.11/ref/templates/builtins/)",
                verbose_name='Indexing expression'
            ),
        )
    ]



from unittest import skip

from django.test import override_settings
from django.utils.module_loading import import_string

from mayan.apps.testing.tests.base import BaseTestCase

from ..decorators import locked_class_method
from ..exceptions import LockError

from .literals import TEST_LOCK_1


class FileLockDecoratorTestCase(BaseTestCase):
    _test_locking_backend_string = 'mayan.apps.lock_manager.backends.file_lock.FileLock'

    def setUp(self):
        super().setUp()
        self.locking_backend = import_string(
            dotted_path=self._test_locking_backend_string
        )

    def test_decorator_single_class(self):
        class TestClass:
            def _lock_manager_get_lock_name(self, *args, **kwargs):
                return TEST_LOCK_1

            @locked_class_method
            def method_1(self, _acquire_lock=True):
                """Locked parent method"""
                self.method_2()

            @locked_class_method
            def method_2(self, _acquire_lock=True):
                """Locked child method"""

        test_object = TestClass()
        with self.assertRaises(expected_exception=LockError):
            test_object.method_1()

    def test_decorator_multiple_classes(self):
        class TestClass1:
            def _lock_manager_get_lock_name(self, *args, **kwargs):
                return TEST_LOCK_1

            @locked_class_method
            def method_1(self, _acquire_lock=True):
                test_object_2 = TestClass2()
                test_object_2.method_1()

        class TestClass2:
            def _lock_manager_get_lock_name(self, *args, **kwargs):
                """Locked parent class method"""
                return TEST_LOCK_1

            @locked_class_method
            def method_1(self, _acquire_lock=True):
                """Locked child class method"""

        test_object_1 = TestClass1()

        with self.assertRaises(expected_exception=LockError):
            test_object_1.method_1()


class ModelLockTestCase(FileLockDecoratorTestCase):
    _test_locking_backend_string = 'mayan.apps.lock_manager.backends.model_lock.ModelLock'


@skip(reason='Skip until a Mock Redis server class is added.')
@override_settings(
    LOCK_MANAGER_BACKEND_ARGUMENTS={'redis_url': 'redis://127.0.0.1:6379/0'}
)
class RedisLockTestCase(FileLockDecoratorTestCase):
    _test_locking_backend_string = 'mayan.apps.lock_manager.backends.redis_lock.RedisLock'



from django.core.management.base import BaseCommand
from django.core.management.utils import get_random_secret_key


class Command(BaseCommand):
    help = 'Generate a random value that is valid for use as a SECRET_KEY.'

    def handle(self, *args, **options):
        self.stdout.write(
            msg=get_random_secret_key()
        )



from distutils import util
from functools import reduce
import itertools
import logging
import types

from django.core.exceptions import FieldDoesNotExist
from django.db.models.constants import LOOKUP_SEP
from django.utils.text import slugify

from .compatibility import Iterable
from .exceptions import ResolverError, ResolverPipelineError

logger = logging.getLogger(name=__name__)


class ProgressBar:
    def __init__(
        self, total, prefix=None, suffix=None,
        decimal_places=1, length=100, fill_symbol='█', print_end='\r'
    ):
        self.total = total
        self.prefix = prefix
        self.suffix = suffix
        self.decimal_places = decimal_places
        self.length = length
        self.fill_symbol = fill_symbol
        self.print_end = print_end
        self.template_percent = '{{0:.{}f}}'.format(self.decimal_places)

    def update(self, index):
        percent = self.template_percent.format(
            index / self.total * 100.0
        )
        fill_size = int(self.length * index // self.total)
        bar = '{}{}'.format(
            self.fill_symbol * fill_size, '-' * (self.length - fill_size)
        )
        print(
            f'\r{self.prefix} |{bar}| {percent}% {self.suffix}',
            end=self.print_end
        )
        # Print New Line on Complete
        if index == self.total:
            print()


class Resolver:
    exceptions = ()

    def __init__(self, attribute, obj, kwargs, klass, resolver_extra_kwargs):
        self.attribute = attribute
        self.obj = obj
        self.kwargs = kwargs
        self.klass = klass
        self.resolver_extra_kwargs = resolver_extra_kwargs

    def resolve(self):
        try:
            return self._resolve()
        except self.exceptions:
            raise ResolverError

    def _resolve(self):
        raise NotImplementedError


class ResolverObjectAttribute(Resolver):
    exceptions = (TypeError,)

    def _resolve(self):
        return self.attribute(self.obj, **self.kwargs)


class ResolverGetattr(Resolver):
    exceptions = (AttributeError, TypeError,)

    def _resolve(self):
        return getattr(self.obj, self.attribute)


class ResolverFunction(Resolver):
    exceptions = (AttributeError, TypeError,)

    def _resolve(self):
        return getattr(self.obj, self.attribute)(**self.kwargs)


class ResolverDictionary(Resolver):
    exceptions = (TypeError,)

    def _resolve(self):
        return self.obj[self.attribute]


class ResolverList(Resolver):
    exceptions = (TypeError,)

    def _resolve(self):
        result = []
        for item in self.obj:
            result.append(
                self.klass.resolve(
                    attribute=self.attribute, kwargs=self.kwargs, obj=item,
                    resolver_extra_kwargs=self.resolver_extra_kwargs
                )
            )

        return result


class ResolverPipelineObjectAttribute:
    resolver_list = (
        ResolverDictionary, ResolverList, ResolverFunction,
        ResolverObjectAttribute, ResolverGetattr
    )

    @classmethod
    def resolve(
        cls, attribute, obj, resolver_extra_kwargs=None, kwargs=None
    ):
        kwargs = kwargs or {}
        resolver_extra_kwargs = resolver_extra_kwargs or {}

        if '.' in attribute:
            attribute_list = attribute.split('.')
        else:
            attribute_list = (attribute,)

        result = obj
        for attribute in attribute_list:
            for resolver in cls.resolver_list:
                try:
                    result = resolver(
                        attribute=attribute, klass=cls, kwargs=kwargs,
                        obj=result,
                        resolver_extra_kwargs=resolver_extra_kwargs
                    ).resolve()
                except ResolverError:
                    """Expected, try the next resolver in the list."""

            if result == obj:
                raise ResolverPipelineError(
                    'Unable to resolve attribute `{attribute}` of object `{obj}`'.format(
                        attribute=attribute, obj=obj
                    )
                )

        return result


class ResolverRelatedManager(Resolver):
    exceptions = (AttributeError, FieldDoesNotExist)

    def _resolve(self):
        model = self.resolver_extra_kwargs.get(
            'model', {}
        )
        exclude = self.resolver_extra_kwargs.get(
            'exclude', {}
        )

        field = self.obj._meta.get_field(field_name=self.attribute)

        if field.many_to_one:
            # Many to one.
            queryset = field.related_model._meta.default_manager.filter(
                **{field.remote_field.name: self.obj.pk}
            )

            if field.related_model == model:
                queryset = queryset.exclude(**exclude)

            return queryset
        elif field.many_to_many:
            # Many to many from the parent side.
            if hasattr(field, 'get_filter_kwargs_for_object'):
                queryset = getattr(self.obj, field.attname)

                if queryset.model == model:
                    queryset = queryset.exclude(**exclude)
                else:
                    queryset = queryset.all()

                return queryset

        # Many to many from the child side.
        # One to many.
        # One to one.
        queryset = field.remote_field.model._meta.default_manager.filter(
            **{field.remote_field.name: self.obj.pk}
        )

        if field.related_model == model:
            queryset = queryset.exclude(**exclude)

        return queryset


class ResolverPipelineModelAttribute(ResolverPipelineObjectAttribute):
    resolver_list = (
        ResolverDictionary, ResolverList, ResolverRelatedManager,
        ResolverFunction, ResolverObjectAttribute, ResolverGetattr
    )

    @classmethod
    def resolve(
        cls, attribute, obj, kwargs=None, resolver_extra_kwargs=None
    ):
        attribute = attribute.replace(LOOKUP_SEP, '.')
        return super().resolve(
            attribute=attribute, kwargs=kwargs, obj=obj,
            resolver_extra_kwargs=resolver_extra_kwargs
        )


def any_to_bool(value):
    if not isinstance(value, bool):
        value = bool(
            util.strtobool(val=value)
        )
    return value


def convert_to_internal_name(value):
    slug = slugify(value=value)
    slug = slug.replace('-', '_')
    return slug


def flatten_list(value):
    if isinstance(value, (str, bytes)):
        yield value
    else:
        for item in value:
            if isinstance(item, Iterable) and not isinstance(item, (str, bytes)):
                yield from flatten_list(value=item)
            else:
                if item is not None:
                    yield item
                else:
                    yield ''


def get_class_full_name(klass):
    return '{klass.__module__}.{klass.__qualname__}'.format(klass=klass)


def get_related_field(model, related_field_name):
    try:
        local_field_name, remaining_field_path = related_field_name.split(
            LOOKUP_SEP, 1
        )
    except ValueError:
        local_field_name = related_field_name
        remaining_field_path = None

    related_field = model._meta.get_field(field_name=local_field_name)

    if remaining_field_path:
        return get_related_field(
            model=related_field.related_model,
            related_field_name=remaining_field_path
        )

    return related_field


def group_iterator(iterable, group_size=None):
    group_size = group_size or 1

    if group_size > 1:
        while True:
            chunk = tuple(
                itertools.islice(iterable, group_size)
            )
            if not chunk:
                break
            yield chunk
    else:
        yield from iterable


def parse_range(range_string):
    for part in range_string.split(','):
        part = part.strip()

        if '-' in part:
            part_range = part.split('-')
            start = int(
                part_range[0].strip()
            )
            stop = int(
                part_range[1].strip()
            )

            if stop > start:
                step = 1
            else:
                step = -1

            yield from range(start, stop + step, step)
        else:
            try:
                value = int(part)
            except ValueError:
                return
            else:
                yield value


def resolve_attribute(attribute, obj, kwargs=None):
    """
    Resolve the attribute of an object. Behaves like the Python REPL but with
    an unified dotted path schema regardless of the attribute type.
    Supports callables, dictionaries, properties, related model fields.
    """
    if not kwargs:
        kwargs = {}

    # Try as a callable
    try:
        return attribute(obj, **kwargs)
    except TypeError:
        # Try as a dictionary
        try:
            return obj[attribute]
        except TypeError:
            try:
                # If there are dots in the attribute name, traverse them
                # to the final attribute
                result = reduce(
                    getattr, attribute.split('.'), obj
                )
                try:
                    # Try it as a method
                    return result(**kwargs)
                except (TypeError, ValueError):
                    # Try it as a property
                    return result
            except AttributeError:
                # Try as a related model field
                if LOOKUP_SEP in attribute:
                    attribute_replaced = attribute.replace(LOOKUP_SEP, '.')
                    return resolve_attribute(
                        attribute=attribute_replaced, kwargs=kwargs, obj=obj
                    )
                else:
                    raise


def return_attrib(obj, attrib, arguments=None):
    if isinstance(attrib, types.FunctionType):
        return attrib(obj)
    elif isinstance(
        obj, dict
    ) or isinstance(obj, dict):
        return obj[attrib]
    else:
        result = reduce(
            getattr, attrib.split('.'), obj
        )
        if isinstance(result, types.MethodType):
            if arguments:
                return result(**arguments)
            else:
                return result()
        else:
            return result


def return_related(instance, related_field):
    """
    This functions works in a similar method to return_attrib but is
    meant for related models. Support multiple levels of relationship
    using double underscore.
    """
    return reduce(
        getattr, related_field.split(LOOKUP_SEP), instance
    )
